# command time python /gale/ddn/snm3C/humanPFC/code/comp_cpg_cell.py --indir /gale/raidix/rdx-5/zhoujt/projects/methylHiC/PFC_batch_merged/smoothed_matrix/1cell/${res0}b_resolution/ --outdir /gale/ddn/snm3C/humanPFC/smoothed_matrix/${res0}b_resolution/ --cell $(cat /gale/ddn/snm3C/humanPFC/smoothed_matrix/celllist_long.txt | head -${SGE_TASK_ID} | tail -1) --chrom ${c} --mode raw --cpg_file /gale/ddn/snm3C/humanPFC/smoothed_matrix/${res0}b_resolution/bins/hg19.${res0}bin.CpG.txt

import h5py
import argparse
import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix, diags

def comp_cpg_cell(indir, outdir, cell, chrom, mode, cpg_file):

	if chrom[:3]=='chr':
		c = chrom[3:]
	else:
		c = chrom
	cpg = pd.read_csv(cpg_file, sep='\t', header=0)
	cpg['ratio'] = (cpg['13_user_patt_count'] / (cpg['12_seq_len'] - cpg['10_num_N'])).fillna(0)
	cpg = cpg[(cpg['#1_usercol']==('chr'+c))]['ratio'].values[:,None]
	binfilter = (cpg[:,0]>0)
	if mode=='raw':
		D = np.loadtxt(f'{indir}chr{c}/{cell}_chr{c}.txt')
		if len(D)==0:
			D = np.array([[0,0,0]])
		elif len(D.shape)==1:
			D = D.reshape(1,-1)
		A = csr_matrix((D[:, 2], (D[:, 0], D[:, 1])), shape = (len(cpg), len(cpg)))
	else:
		with h5py.File(f'{indir}chr{c}/{cell}_chr{c}_{mode}.hdf5', 'r') as f:
			g = f['Matrix']
			A = csr_matrix((g['data'][()], g['indices'][()], g['indptr'][()]), g.attrs['shape'])
	A = A - diags(A.diagonal())
	A = A + A.T
	A = A + diags((A.sum(axis=0).A.ravel()==0).astype(int))
	# normalize row sum to 1
	d = diags(1 / A.sum(axis=0).A.ravel())
	P = d.dot(A)
	# weighted average of cpg density
	comp = A.dot(cpg)[:,0]
	tmp = comp[binfilter]
	Apos = (tmp > np.percentile(tmp, 80))
	Bpos = (tmp < np.percentile(tmp, 20))
	E = A.tocoo()
	decay = np.array([E.diagonal(i).mean() for i in range(E.shape[0])])
	E.data = E.data / decay[np.abs(E.col - E.row)]
	E = E.tocsr()[np.ix_(binfilter, binfilter)]
	score = [E[np.ix_(Apos, Apos)].sum(), E[np.ix_(Bpos, Bpos)].sum(), E[np.ix_(Apos, Bpos)].sum()]
	np.save(f'{outdir}chr{c}/{cell}_chr{c}_{mode}.cpgcomp.npy', comp.tolist() + score)
	return

'''
parser = argparse.ArgumentParser()
parser.add_argument('--indir', type=str, default=None, help='Directory of raw/imputed matrices end with /')
parser.add_argument('--outdir', type=str, default=None, help='Directory of compartment matrix end with /')
parser.add_argument('--cell', type=str, default=None, help='Full path of a file containing a list of cell identifiers to be concatenate')
parser.add_argument('--chrom', type=str, default=None, help='Chromosome to impute')
parser.add_argument('--mode', type=str, default=None, help='Suffix of imputed matrix file names')
parser.add_argument('--cpg_file', type=str, default=None, help='CpG counts generated by bedtools nuc')
opt = parser.parse_args()

comp_cpg_cell(opt.indir, opt.outdir, opt.cell, opt.chrom, opt.mode, opt.cpg_file)
'''



# command time python /gale/ddn/snm3C/humanPFC/code/loop_bkg_cell.py --indir /gale/ddn/snm3C/humanPFC/smoothed_matrix/${res0}b_resolution/ --cell ${sample} --chrom ${c} --res ${res} --impute_mode pad2_std1_rp0.5_ws40

import sys
import h5py
import time
import cv2
cv2.useOptimized()
import argparse
import numpy as np
from scipy.sparse import save_npz, csr_matrix
from sklearn.preprocessing import RobustScaler

def loop_bkg_cell(indir, cell, chrom, impute_mode, res, 
			dist=10050000, cap=5, pad=5, gap=2, norm_mode='dist_trim'):

	if chrom[:3]=='chr':
		c = chrom
	else:
		c = 'chr' + chrom

	start_time = time.time()
	with h5py.File(f'{indir}{c}/{cell}_{c}_{impute_mode}.hdf5', 'r') as f:
		g = f['Matrix']
		E = csr_matrix((g['data'][()], g['indices'][()], g['indptr'][()]), g.attrs['shape']).tocoo()

	print('Load', time.time() - start_time)

	# TODO numba optimize
	start_time = time.time()
	ave, std, top, count = np.zeros((4, dist // res + pad + 1))
	for i in range(dist // res + pad + 1):
		tmp = E.diagonal(i)
		top[i] = np.percentile(tmp,99)
		tmp[tmp>top[i]] = top[i]
		ave[i] = np.mean(tmp)
		std[i] = np.std(tmp)
		count[i] = np.sum(tmp>0)

	print('Curve', time.time() - start_time, '#Nonzero', np.sum(count))

	idx = np.triu_indices(E.shape[0], 0)
	idxfilter = ((idx[1] - idx[0]) < (dist // res + 1))
	idx = (idx[0][idxfilter], idx[1][idxfilter])
	mask = csr_matrix((np.ones(len(idx[0])), (idx[0], idx[1])), E.shape)

	start_time = time.time()
	E.data = np.min([E.data, top[E.col - E.row]], axis=0)
	E = E.astype(np.float32).toarray()
	tmp = E[idx]
	tmp = (tmp - ave[idx[1] - idx[0]]) / std[idx[1] - idx[0]]
	tmp[count[idx[1] - idx[0]] < 100] = 0
	tmp[std[idx[1] - idx[0]]==0] = 0
	tmp[tmp > cap] = cap
	tmp[tmp < -cap] = -cap
	E[idx] = tmp.copy()
	print('Norm', time.time() - start_time)

	start_time = time.time()
	w = pad * 2 + 1
	kernel = np.ones((w, w), np.float32)
	kernel[(pad - gap):(pad + gap + 1), (pad - gap):(pad + gap + 1)] = 0
	kernel = kernel / np.sum(kernel)
	N = cv2.filter2D(E, -1, kernel=kernel)
	E = csr_matrix(np.around(E, decimals=6))
	N = csr_matrix(np.around(N, decimals=6)).multiply(mask)
	N = E - N
	print('Bkg', time.time() - start_time)

	save_npz(f'{indir}{c}/{cell}_{c}_{impute_mode}_{norm_mode}.E.npz', E)
	save_npz(f'{indir}{c}/{cell}_{c}_{impute_mode}_{norm_mode}.T.npz', N)
	return

'''
parser = argparse.ArgumentParser()
parser.add_argument('--indir', type=str, default=None, help='Directory of imputed matrix')
parser.add_argument('--cell', type=str, default=None, help='Specific identifier of a cell')
parser.add_argument('--chrom', type=str, default=None, help='Chromosome imputed')
parser.add_argument('--impute_mode', type=str, default=None, help='Suffix of imputed matrix file names')
parser.add_argument('--res', type=int, default=None, help='Bin size as integer to generate contact matrix')

parser.add_argument('--dist', type=int, default=10050000, help='Maximum distance threshold of contacts to use')
parser.add_argument('--cap', type=int, default=5, help='Trim Z-scores over the threshold')
parser.add_argument('--pad', type=int, default=5, help='One direction size of larger square for donut background')
parser.add_argument('--gap', type=int, default=2, help='One direction size of smaller square for donut background')
parser.add_argument('--norm_mode', type=str, default='dist_trim', help='Suffix of normalized file names')
opt = parser.parse_args()

loop_bkg_cell(opt.indir, opt.cell, opt.chrom, opt.impute_mode, opt.res, 
		opt.dist, opt.cap, opt.pad, opt.gap, opt.norm_mode)
'''



import time
import cv2

cv2.useOptimized()
import numpy as np
from scipy.sparse import load_npz, save_npz, csr_matrix


def loop_sc(outdir, cell, chrom, impute_mode, res, dist, cap, pad, gap, norm_mode):
    if chrom[:3] == 'chr':
        c = chrom[3:]
    else:
        c = chrom

    E = load_npz(outdir + cell + '_chr' + c + '_' + impute_mode + '.npz')
    start_time = time.time()
    ave, std, top, count = np.zeros((4, dist // res + pad + 1))
    for i in range(dist // res + pad + 1):
        tmp = E.diagonal(i)
        top[i] = np.percentile(tmp, 99)
        tmp[tmp > top[i]] = top[i]
        ave[i] = np.mean(tmp)
        std[i] = np.std(tmp)
        count[i] = np.sum(tmp > 0)

    print('Curve', time.time() - start_time, '#Nonzero', np.sum(count))
    start_time = time.time()
    E.data = np.min([E.data, top[E.col - E.row]], axis=0)
    E = E.astype(np.float32).toarray()

    idx = np.triu_indices(E.shape[0], 0)
    idxfilter = ((idx[1] - idx[0]) < (dist // res + 1))
    idx = (idx[0][idxfilter], idx[1][idxfilter])
    mask = csr_matrix((np.ones(len(idx[0])), (idx[0], idx[1])), E.shape)

    tmp = E[idx]
    tmp = (tmp - ave[idx[1] - idx[0]]) / std[idx[1] - idx[0]]
    tmp[count[idx[1] - idx[0]] < 100] = 0
    tmp[std[idx[1] - idx[0]] == 0] = 0
    tmp[tmp > cap] = cap
    tmp[tmp < -cap] = -cap
    E[idx] = tmp.copy()
    print('Norm', time.time() - start_time)

    start_time = time.time()
    w = pad * 2 + 1
    kernel = np.ones((w, w), np.float32)
    kernel[(pad - gap):(pad + gap + 1), (pad - gap):(pad + gap + 1)] = 0
    kernel = kernel / np.sum(kernel)
    N = cv2.filter2D(E, -1, kernel=kernel)
    E = csr_matrix(np.around(E, decimals=6))
    N = csr_matrix(np.around(N, decimals=6)).multiply(mask)
    N = E - N
    print('Bkg', time.time() - start_time)

    save_npz(outdir + cell + '_chr' + c + '_' + impute_mode + '_' + norm_mode + '.E.npz', E)
    save_npz(outdir + cell + '_chr' + c + '_' + impute_mode + '_' + norm_mode + '.T.npz', N)
    return



import time
import numpy as np
from schicluster import *
from sklearn.cluster import KMeans, SpectralClustering
from sklearn.metrics.cluster import adjusted_rand_score as ARI

mm9dim = [197195432, 181748087, 159599783, 155630120, 152537259, 149517037, 152524553, 131738871, 124076172, 129993255,
          121843856, 121257530, 120284312, 125194864, 103494974, 98319150, 95272651, 90772031, 61342430]
hg19dim = [249250621, 243199373, 198022430, 191154276, 180915260, 171115067, 159138663, 146364022, 141213431, 135534747,
           135006516, 133851895, 115169878, 107349540, 102531392, 90354753, 81195210, 78077248, 59128983, 63025520,
           48129895, 51304566, 155270560]

# File list and labels of dataset Ramani 2017
ctlist = ['HeLa', 'HAP1', 'GM12878', 'K562']
network = [np.loadtxt('1mb_resolution/' + ct + '/sample_list.txt', dtype=np.str) for ct in ctlist]
label = np.array([ctlist[i] for i in range(len(ctlist)) for j in range(len(network[i]))]).astype('U8')
network = np.concatenate(network)
chrom = [str(i + 1) for i in range(22)] + ['X']
chromsize = {chrom[i]: hg19dim[i] for i in range(len(chrom))}
nc = 4

# CpG content for each bin
cg = np.loadtxt('hg19/bin/hg19.1mb.bin.CpG.txt', dtype=np.str, skiprows=1, usecols=(0, 9, 11, 12))
cgdata = cg[:, 1:].astype(float)
cgdata = cgdata[:, 2] / (cgdata[:, 1] - cgdata[:, 0])
cgdata[np.isnan(cgdata)] = 0.0
chrcg = {c: cgdata[cg[:, 0] == 'chr' + c] for c in chrom}

# scHiCluster GPU
start_time = time.time()
cluster, embedding = hicluster_gpu(network, chromsize, nc=nc)
print(time.time() - start_time)
[ARI(label, KMeans(n_clusters=nc, n_init=200).fit(embedding[:, :ndim]).labels_) for ndim in [2, 5, 10, 20, 50]]

# scHiCluster CPU
start_time = time.time()
cluster, embedding = hicluster_cpu(network, chromsize, nc=nc, ncpus=5)
print(time.time() - start_time)
[ARI(label, KMeans(n_clusters=nc, n_init=200).fit(embedding[:, :ndim]).labels_) for ndim in [2, 5, 10, 20, 50]]

# PCA
start_time = time.time()
cluster, embedding = raw_pca(network, chromsize, nc=nc)
print(time.time() - start_time)
[ARI(label, KMeans(n_clusters=nc, n_init=200).fit(embedding[:, 1:(ndim + 1)]).labels_) for ndim in [2, 5, 10, 20, 50]]
[ARI(label, SpectralClustering(n_clusters=nc, affinity='nearest_neighbors').fit(embedding[:, 1:(ndim + 1)]).labels_) for
 ndim in [2, 5, 10, 20, 50]]

# Downsample reads to uniform the coverage of all the cells before PCA
start_time = time.time()
cluster, embedding = ds_pca(network, chromsize, nc=nc)
print(time.time() - start_time)
[ARI(label, KMeans(n_clusters=nc, n_init=200).fit(embedding[:, :ndim]).labels_) for ndim in [2, 5, 10, 20, 50]]
[ARI(label, SpectralClustering(n_clusters=nc, affinity='nearest_neighbors').fit(embedding[:, :ndim]).labels_) for ndim
 in [2, 5, 10, 20, 50]]

# Use compartment score (PC1) of single cells
start_time = time.time()
cluster, embedding = compartment(network, chromsize, nc=nc)
print(time.time() - start_time)
[ARI(label, KMeans(n_clusters=nc, n_init=200).fit(embedding[:, :ndim]).labels_) for ndim in [2, 5, 10, 20, 50]]
[ARI(label, SpectralClustering(n_clusters=nc, affinity='nearest_neighbors').fit(embedding[:, :ndim]).labels_) for ndim
 in [2, 5, 10, 20, 50]]

# Use contact-distance decay curve
start_time = time.time()
cluster, embedding = decay(network, chromsize, nc=nc)
print(time.time() - start_time)
[ARI(label, KMeans(n_clusters=nc, n_init=200).fit(embedding[:, :ndim]).labels_) for ndim in [2, 5, 10, 20, 50]]
[ARI(label, SpectralClustering(n_clusters=nc, affinity='nearest_neighbors').fit(embedding[:, :ndim]).labels_) for ndim
 in [2, 5, 10, 20, 50]]

# scHiCluster without linear convolution
start_time = time.time()
cluster, embedding = hicluster_gpu(network, chromsize, nc=nc, pad=0)
print(time.time() - start_time)
[ARI(label, KMeans(n_clusters=nc, n_init=200).fit(embedding[:, :ndim]).labels_) for ndim in [2, 5, 10, 20, 50]]
[ARI(label, SpectralClustering(n_clusters=nc, affinity='nearest_neighbors').fit(embedding[:, :ndim]).labels_) for ndim
 in [2, 5, 10, 20, 50]]

# scHiCluster without random walk
start_time = time.time()
cluster, embedding = hicluster_gpu(network, chromsize, nc=nc, rp=-1)
print(time.time() - start_time)
[ARI(label, KMeans(n_clusters=nc, n_init=200).fit(embedding[:, :ndim]).labels_) for ndim in [2, 5, 10, 20, 50]]
[ARI(label, SpectralClustering(n_clusters=nc, affinity='nearest_neighbors').fit(embedding[:, :ndim]).labels_) for ndim
 in [2, 5, 10, 20, 50]]

# scHiCluster without keeping the top elements
start_time = time.time()
cluster, embedding = hicluster_gpu(network, chromsize, nc=nc, prct=-1)
print(time.time() - start_time)
[ARI(label, KMeans(n_clusters=nc, n_init=200).fit(embedding[:, 1:(ndim + 1)]).labels_) for ndim in [2, 5, 10, 20, 50]]
[ARI(label, SpectralClustering(n_clusters=nc, affinity='nearest_neighbors').fit(embedding[:, 1:(ndim + 1)]).labels_) for
 ndim in [2, 5, 10, 20, 50]]
np.save('/cellar/users/zhoujt1994/projects/scHiC/' + dataset + '/embedding/1mb_pad1_rwr_real.npy', embedding)

# Random walk only
start_time = time.time()
cluster, embedding = hicluster_gpu(network, chromsize, nc=nc, pad=0, prct=-1)
print(time.time() - start_time)
[ARI(label, KMeans(n_clusters=nc, n_init=200).fit(embedding[:, 1:(ndim + 1)]).labels_) for ndim in [2, 5, 10, 20, 50]]
[ARI(label, SpectralClustering(n_clusters=nc, affinity='nearest_neighbors').fit(embedding[:, 1:(ndim + 1)]).labels_) for
 ndim in [2, 5, 10, 20, 50]]

# Linear convolution only
start_time = time.time()
cluster, embedding = hicluster_gpu(network, chromsize, nc=nc, rp=-1, prct=-1)
print(time.time() - start_time)
[ARI(label, KMeans(n_clusters=nc, n_init=200).fit(embedding[:, 1:(ndim + 1)]).labels_) for ndim in [2, 5, 10, 20, 50]]
[ARI(label, SpectralClustering(n_clusters=nc, affinity='nearest_neighbors').fit(embedding[:, 1:(ndim + 1)]).labels_) for
 ndim in [2, 5, 10, 20, 50]]



import numpy as np
import pandas as pd
import cooler
from sklearn.decomposition import TruncatedSVD
import joblib
import subprocess
from concurrent.futures import ProcessPoolExecutor, as_completed
import pathlib


def make_idx(n_dim, dist, resolution):
    idx = np.triu_indices(n_dim, k=1)
    idx_filter = np.array([(yy - xx) < (dist / resolution + 1)
                           for xx, yy in zip(idx[0], idx[1])])
    idx = (idx[0][idx_filter], idx[1][idx_filter])
    return idx


def make_chrom_matrix(cell_table,
                      chrom,
                      nbins,
                      output_path,
                      scale_factor,
                      dist,
                      resolution):
    idx = make_idx(nbins, dist, resolution)
    shape = (cell_table.size, idx[0].size)
    # read data
    chrom_matrix = np.zeros(shape=shape, dtype='float32')
    for i, (_, cell_url) in enumerate(cell_table.items()):
        cool = cooler.Cooler(cell_url)
        matrix = cool.matrix(balance=False, sparse=False).fetch(chrom)
        # each row of chrom_matrix is a 1D cell-chrom matrix
        chrom_matrix[i, :] = matrix[idx].ravel()
    chrom_matrix *= scale_factor
    np.savez(output_path, chrom_matrix)
    return


def svd(input_path, dim, output_prefix, save_model=True, norm_sig=True):
    chrom_matrix = np.load(input_path)['arr_0']
    dim = min(dim, chrom_matrix.shape[0] - 1, chrom_matrix.shape[1] - 1)
    model = TruncatedSVD(n_components=dim, algorithm='arpack')
    decomp = model.fit_transform(chrom_matrix)

    if norm_sig:
        decomp = decomp[:, model.singular_values_ > 0]
        singular_values = model.singular_values_[model.singular_values_ > 0]
        decomp /= singular_values[None, :]

    np.savez(f'{output_prefix}_decomp.npz', decomp)
    if save_model:
        joblib.dump(model, f'{output_prefix}_SVD.lib')
    return


def embedding(cell_table_path,
              output_dir,
              dim=50,
              dist=1000000,
              resolution=100000,
              scale_factor=100000,
              norm_sig=True,
              cpu=1,
              save_model=False,
              save_raw=True):
    cell_table = pd.read_csv(cell_table_path,
                             sep='\t',
                             index_col=0,
                             header=None,
                             squeeze=True)

    output_dir = pathlib.Path(output_dir).absolute()
    raw_dir = output_dir / 'raw'
    raw_dir.mkdir(exist_ok=True)

    first_cool = cooler.Cooler(cell_table.iloc[0])
    chroms = first_cool.chromnames
    chrom_bin_counts = first_cool.bins()[:]['chrom'].value_counts()
    # remove small chroms
    chroms = [chrom for chrom in chroms if chrom_bin_counts[chrom] > 2]

    # prepare raw chromosome 1D matrix
    with ProcessPoolExecutor(cpu) as exe:
        futures = {}
        for chrom in chroms:
            nbins = chrom_bin_counts[chrom]
            output_path = raw_dir / f'{chrom}.npz'
            future = exe.submit(make_chrom_matrix,
                                cell_table=cell_table,
                                chrom=chrom,
                                nbins=nbins,
                                output_path=output_path,
                                scale_factor=scale_factor,
                                dist=dist,
                                resolution=resolution)
            futures[future] = chrom

        for future in as_completed(futures):
            chrom = futures[future]
            print(f'{chrom} matrix generated')
            future.result()

    # SVD on each chromosome
    decomp_dir = output_dir / 'decomp'
    decomp_dir.mkdir(exist_ok=True)
    with ProcessPoolExecutor(cpu) as exe:
        futures = {}
        for chrom in chroms:
            chrom_raw_path = raw_dir / f'{chrom}.npz'
            output_prefix = decomp_dir / chrom
            future = exe.submit(svd,
                                input_path=chrom_raw_path,
                                dim=dim,
                                output_prefix=output_prefix,
                                save_model=save_model,
                                norm_sig=norm_sig)
            futures[future] = chrom

        for future in as_completed(futures):
            chrom = futures[future]
            print(f'{chrom} SVD generated')
            future.result()

    # concatenate all chromosome decomp matrix and do another SVD
    total_data = []
    for chrom in chroms:
        decomp_path = decomp_dir / f'{chrom}_decomp.npz'
        data = np.load(str(decomp_path))['arr_0']
        total_data.append(data)
    total_data = np.concatenate(total_data, axis=1)
    total_data_path = decomp_dir / f'total_chrom_decomp_concat.npz'
    np.savez(total_data_path, total_data)

    # final SVD
    output_prefix = decomp_dir / f'total'
    svd(input_path=total_data_path,
        dim=dim,
        output_prefix=output_prefix,
        save_model=save_model,
        norm_sig=norm_sig)

    # clean single chrom
    for chrom in chroms:
        decomp_path = decomp_dir / f'{chrom}_decomp.npz'
        subprocess.run(['rm', '-f', str(decomp_path)])
    if not save_raw:
        subprocess.run(['rm', '-rf', str(raw_dir)])
    return



import time
import cv2
import argparse
import numpy as np
import pandas as pd
from scipy import stats
from scipy.sparse import save_npz, load_npz, csr_matrix, coo_matrix

parser = argparse.ArgumentParser()


def merge_cell(indir,
               cell_list,
               group,
               chrom,
               res,
               impute_mode,
               norm_mode='dist_trim',
               min_dist=50000,
               max_dist=10000000,
               pad=5,
               gap=2):

    if chrom[:3] == 'chr':
        c = chrom[3:]
    else:
        c = chrom

    thres = stats.norm(0, 1).isf(0.025)
    celllist = np.loadtxt(indir + 'merged/' + cell_list, dtype=np.str)
    tot = len(celllist)
    Q = load_npz(indir + 'chr' + c + '/' + celllist[0] + '_chr' + c + '_' + impute_mode + '.npz')
    Qsum, Esum, Osum, Nsum, N2sum = [csr_matrix(Q.shape) for i in range(5)]
    start_time = time.time()
    for i, cell in enumerate(celllist):
        Q = load_npz(indir + 'chr' + c + '/' + cell + '_chr' + c + '_' + impute_mode + '.npz')
        E = load_npz(
            indir + 'chr' + c + '/' + cell + '_chr' + c + '_' + impute_mode + '_' + norm_mode + '.E.npz')
        O = E.copy()
        O.data = (O.data > thres).astype(int)
        Qsum += Q
        Osum += O
        Esum += E

    Qsum.data = Qsum.data / tot
    Esum.data = Esum.data / tot
    Osum.data = Osum.data / tot

    save_npz(indir + 'merged/' + group + '_' + impute_mode + '_' + norm_mode + '.chr' + c + '.Q.npz',
             Qsum)
    save_npz(indir + 'merged/' + group + '_' + impute_mode + '_' + norm_mode + '.chr' + c + '.E.npz',
             Esum)
    save_npz(indir + 'merged/' + group + '_' + impute_mode + '_' + norm_mode + '.chr' + c + '.O.npz',
             Osum)
    print('Merge cell', time.time() - start_time)

    E = Esum.toarray()
    O = Osum.toarray()
    del Qsum, Esum, Osum

    start_time = time.time()
    oefilter = np.logical_and(E > 0, O > 0.1)
    loop = np.where(oefilter)
    distfilter = np.logical_and((loop[1] - loop[0]) > (min_dist / res),
                                (loop[1] - loop[0]) < (max_dist / res))
    loop = (loop[0][distfilter], loop[1][distfilter])

    start_time = time.time()
    eloop = np.zeros((tot, len(loop[0])))
    for i, cell in enumerate(celllist):
        eloop[i] = load_npz(
            indir + 'chr' + c + '/' + cell + '_chr' + c + '_' + impute_mode + '_' + norm_mode + '.T.npz')[
            loop].A.ravel()

    print('Load loop', time.time() - start_time)

    pvr = np.array([stats.wilcoxon(xx, alternative='greater')[1] for xx in eloop.T])
    pvt = stats.ttest_1samp(eloop, 0, axis=0)
    pvt[1][pvt[0] > 0] *= 2
    pvt[1][pvt[0] <= 0] = 1
    pvt = pvt[1]
    print('Test loop', time.time() - start_time)
    del eloop

    w = pad * 2 + 1
    start_time = time.time()

    kernel_bl = np.zeros((w, w), np.float32)
    kernel_bl[-pad:, :(pad - gap)] = 1
    kernel_bl[-(pad - gap):, :pad] = 1

    kernel_donut = np.ones((w, w), np.float32)
    kernel_donut[pad, :] = 0
    kernel_donut[:, pad] = 0
    kernel_donut[(pad - gap):(pad + gap + 1), (pad - gap):(pad + gap + 1)] = 0

    kernel_lr = np.ones((3, w), np.float32)
    kernel_lr[:, (pad - gap):(pad + gap + 1)] = 0

    kernel_bu = np.ones((w, 3), np.float32)
    kernel_bu[(pad - gap):(pad + gap + 1), :] = 0

    kernel_bl = kernel_bl / np.sum(kernel_bl)
    kernel_donut = kernel_donut / np.sum(kernel_donut)
    kernel_lr = kernel_lr / np.sum(kernel_lr)
    kernel_bu = kernel_bu / np.sum(kernel_bu)

    Ebl = cv2.filter2D(E, -1, kernel=kernel_bl) * (E > 0)
    Edonut = cv2.filter2D(E, -1, kernel=kernel_donut) * (E > 0)
    Elr = cv2.filter2D(E, -1, kernel=kernel_lr) * (E > 0)
    Ebu = cv2.filter2D(E, -1, kernel=kernel_bu) * (E > 0)

    data = np.array([loop[0], loop[1], pvr, pvt, (E/Ebl)[loop], (E/Edonut)[loop], (E/Elr)[loop], (E/Ebu)[loop]]).T
    np.save(indir + 'merged/' + group + '_' + impute_mode + '_' + norm_mode + '.chr' + c + '.loop.npy',
            data)
    print('Filter loop', time.time() - start_time)

    return


'''
start_time = time.time()
E = coo_matrix(E)
data = np.array([np.zeros(len(E.data)).astype(int), np.repeat(['chr'+c], len(E.data)), 
E.row*res, np.zeros(len(E.data)).astype(int), np.ones(len(E.data)).astype(int), 
np.repeat(['chr'+c], len(E.data)), E.col*res, np.ones(len(E.data)).astype(int), np.around(E.data, decimals=2)]).T
data = pd.DataFrame(data, columns=['str1', 'chr1', 'x1', 'frag1','str2', 'chr2', 'y1', 'frag2', 'score'])
data.to_csv(indir + 'merged/' + ct + '_pad2_std1_rp0.5_sqrtvc_distnz_trim5.chr' + c + '.E.txt.gz', 
index=False, header=None, sep='\t', compression='gzip')
print('Write E', time.time() - start_time)
'''



import cooler
import numpy as np
from scipy import stats
from scipy.ndimage import convolve
import pandas as pd
import time
from statsmodels.stats.multitest import multipletests
from heapq import heappop, heapify


def fetch_chrom(cool, chrom) -> np.array:
    return cool.matrix(balance=False, sparse=True).fetch(chrom).toarray()


def select_loop_candidates(cool_e, min_dist, max_dist, resolution, chrom):
    """Select loop candidate pixel to perform t test"""
    E = fetch_chrom(cool_e, chrom)
    loop = np.where(E > 0)  # loop is [xs, ys] of E

    # only calculate upper triangle and remove the pixels close to diagonal
    dist_filter = np.logical_and((loop[1] - loop[0]) > (min_dist / resolution),
                                 (loop[1] - loop[0]) < (max_dist / resolution))
    loop = (loop[0][dist_filter], loop[1][dist_filter])
    print(dist_filter.sum(), 'loop candidate pixels')
    return E, loop


def paired_t_test(cool_t, cool_t2, chrom, loop, n_cells):
    """Paired t test per pixel"""
    T = fetch_chrom(cool_t, chrom)
    T2 = fetch_chrom(cool_t2, chrom)

    loop_delta = T[loop]
    loop_t = loop_delta * n_cells
    loop_t2 = T2[loop] * n_cells
    sed = np.sqrt((loop_t2 - loop_t ** 2 / n_cells) / (n_cells - 1) / n_cells)
    t_score = loop_delta / sed
    p_value = stats.t.sf(t_score, n_cells - 1)
    # Cohen d effect size
    d = 2 * t_score / np.sqrt(2 * n_cells)
    # TODO permute FDR
    return p_value, loop_delta, d


def scan_kernel(E, kernel, loop):
    """Scan loop surrounding background kernel"""
    E_kernel = convolve(E, kernel, mode='mirror') * (E > 0)
    return E_kernel[loop]


def loop_background(E, pad, gap, loop):
    """Calculate loop surrounding background level"""
    w = pad * 2 + 1
    kernel_bl = np.zeros((w, w), np.float32)
    kernel_bl[-pad:, :(pad - gap)] = 1
    kernel_bl[-(pad - gap):, :pad] = 1
    kernel_bl = kernel_bl / np.sum(kernel_bl)
    loop_bl = scan_kernel(E, kernel_bl, loop)

    kernel_donut = np.ones((w, w), np.float32)
    kernel_donut[pad, :] = 0
    kernel_donut[:, pad] = 0
    kernel_donut[(pad - gap):(pad + gap + 1), (pad - gap):(pad + gap + 1)] = 0
    kernel_donut = kernel_donut / np.sum(kernel_donut)
    loop_donut = scan_kernel(E, kernel_donut, loop)

    kernel_h = np.ones((3, w), np.float32)
    kernel_h[:, (pad - gap):(pad + gap + 1)] = 0
    kernel_h = kernel_h / np.sum(kernel_h)
    loop_h = scan_kernel(E, kernel_h, loop)

    kernel_v = np.ones((w, 3), np.float32)
    kernel_v[(pad - gap):(pad + gap + 1), :] = 0
    kernel_v = kernel_v / np.sum(kernel_v)
    loop_v = scan_kernel(E, kernel_v, loop)
    return loop_bl, loop_donut, loop_h, loop_v


def call_loop_single_chrom(group_prefix,
                           chrom,
                           resolution=10000,
                           min_dist=50000,
                           max_dist=10000000,
                           pad=5,
                           gap=2):
    """calculate t test and loop background for one chromosome"""
    # matrix cool obj
    cool_e = cooler.Cooler(f'{group_prefix}.E.cool')
    cool_e2 = cooler.Cooler(f'{group_prefix}.E2.cool')
    cool_t = cooler.Cooler(f'{group_prefix}.T.cool')
    cool_t2 = cooler.Cooler(f'{group_prefix}.T2.cool')
    n_cells = cool_e.info['group_n_cells']

    # call loop
    print(f'{chrom}\tSelecting loop candidates.')
    E, loop = select_loop_candidates(cool_e=cool_e,
                                     min_dist=min_dist,
                                     max_dist=max_dist,
                                     resolution=resolution,
                                     chrom=chrom)

    print(f'{chrom}\tDoing paired T test with the local background.')
    local_p_value, loop_t, local_d = paired_t_test(cool_t=cool_t,
                                                   cool_t2=cool_t2,
                                                   chrom=chrom,
                                                   loop=loop,
                                                   n_cells=n_cells)

    print(f'{chrom}\tDoing paired T test with the global background.')
    global_p_value, loop_e, global_d = paired_t_test(cool_t=cool_e,
                                                     cool_t2=cool_e2,
                                                     chrom=chrom,
                                                     loop=loop,
                                                     n_cells=n_cells)

    print(f'{chrom}\tCalculating loop local background with different masks.')
    loop_bl, loop_donut, loop_h, loop_v = loop_background(E=E,
                                                          pad=pad,
                                                          gap=gap,
                                                          loop=loop)
    # put together loop table
    data = pd.DataFrame({
        'x': loop[0],
        'y': loop[1],
        'distance': (loop[1] - loop[0]) * resolution,
        'local_pval': local_p_value,
        'local_cohen_d': local_d,
        'global_pval': global_p_value,
        'global_cohen_d': global_d,
        'E': loop_e,
        'T': loop_t,
        'E_bl': loop_bl,
        'E_donut': loop_donut,
        'E_h': loop_h,
        'E_v': loop_v
    })
    data['chrom'] = chrom
    return data


def filter_by_background(data, thres_bl, thres_donut, thres_h, thres_v,
                         resolution):
    data['bkfilter'] = (((data['E'] / data['E_bl'] > thres_bl) |
                         (data['E_bl'] < 0)) &
                        ((data['E'] / data['E_donut'] > thres_donut) |
                         (data['E_donut'] < 0)) &
                        ((data['E'] / data['E_h'] > thres_h) |
                         (data['E_h'] < 0)) &
                        ((data['E'] / data['E_v'] > thres_v) |
                         (data['E_v'] < 0)))
    data['x1'] = data['x'].astype(int) * resolution
    data['y1'] = data['y'].astype(int) * resolution
    data['x2'] = data['x1'] + resolution
    data['y2'] = data['y1'] + resolution
    del data['x']
    del data['y']
    return data


def find_summit(loop, res, dist_thres):
    loop = loop.copy()
    start_time = time.time()
    cord = loop[['x1', 'y1']].values // res
    idx = np.argsort(cord[:, 0])
    neighbor = {i: [] for i in range(len(idx))}
    for i in range(len(idx) - 1):
        tmp = cord[idx[i]]
        for j in range(i + 1, len(idx)):
            if cord[idx[j], 0] - tmp[0] > dist_thres:
                break
            if np.abs(tmp[1] - cord[idx[j], 1]) <= dist_thres:
                neighbor[idx[i]].append(idx[j])
                neighbor[idx[j]].append(idx[i])
    # print('Build graph takes', time.time() - start_time, 'seconds')

    start_time = time.time()
    nodescore = loop['E'].values
    flag = np.zeros(len(nodescore))
    tot = len(nodescore)
    summit = []
    nodeheap = (loop['E'] *
                -1).reset_index().reset_index()[['E',
                                                 'level_0']].values.tolist()
    heapify(nodeheap)

    while tot > 0:
        t = int(heappop(nodeheap)[1])
        while flag[t]:
            t = int(heappop(nodeheap)[1])
        q = [t]
        flag[t] = 1
        tot -= 1
        head = 0
        flagtmp = np.zeros(len(nodescore))
        while (head < len(q)):
            for t in neighbor[q[head]]:
                if not flagtmp[t] and nodescore[t] < nodescore[q[head]]:
                    if not flag[t]:
                        flag[t] = 1
                        tot -= 1
                    flagtmp[t] = 1
                    q.append(t)
            head += 1
        summit.append([q[0], len(q)])
    summit = np.array(summit)
    loop = loop.iloc[summit[:, 0]]
    loop['size'] = summit[:, 1]
    # print('BFS takes', time.time() - start_time, 'seconds')
    return loop


def call_loops(group_prefix,
               resolution,
               output_prefix,
               thres_bl=1.33,
               thres_donut=1.33,
               thres_h=1.2,
               thres_v=1.2,
               fdr_thres=0.1,
               dist_thres=20000,
               size_thres=1):
    try:
        group_q = f'{group_prefix}.Q.cool'
        chroms = cooler.Cooler(group_q).chromnames
    except OSError:
        group_q = f'{group_prefix}.Q.mcool::/resolutions/10000'
        chroms = cooler.Cooler(group_q).chromnames
    total_loops = []
    for chrom in chroms:
        print(f'Calling loops of chromosome {chrom}')
        data = call_loop_single_chrom(group_prefix,
                                      chrom,
                                      resolution=10000,
                                      min_dist=50000,
                                      max_dist=10000000,
                                      pad=5,
                                      gap=2)
        total_loops.append(data)
    total_loops = pd.concat(total_loops).reset_index(drop=True)

    # add background judge info
    print('Filtering loop by background.')
    total_loops = filter_by_background(data=total_loops,
                                       thres_bl=thres_bl,
                                       thres_donut=thres_donut,
                                       thres_h=thres_h,
                                       thres_v=thres_v,
                                       resolution=resolution)

    # Group the loops by distance then calculate FDR separately
    print('Filtering loop by FDR.')
    total_loops.dropna(subset=['local_pval', 'global_pval'],
                       how='any',
                       inplace=True)
    local_qs = []
    global_qs = []
    for dist in total_loops['distance'].unique():
        p_values = total_loops.loc[total_loops['distance'] == dist,
                                   ['local_pval', 'global_pval']]
        _, local_q, *_ = multipletests(p_values['local_pval'])
        local_qs.append(pd.Series(local_q, index=p_values.index))
        _, global_q, *_ = multipletests(p_values['global_pval'])
        global_qs.append(pd.Series(global_q, index=p_values.index))
    local_qs = pd.concat(local_qs).sort_index()
    global_qs = pd.concat(global_qs).sort_index()
    total_loops['local_qval'] = local_qs
    total_loops['global_qval'] = global_qs

    # apply all the filters
    loop = total_loops.loc[total_loops['bkfilter']
                           & (total_loops['local_qval'] < fdr_thres)
                           & (total_loops['global_qval'] < fdr_thres)].copy()
    loop.to_hdf(f'{output_prefix}.loop_info.hdf', key='data')

    # filter and save bedpe
    bedpe_cols = ['chrom', 'x1', 'x2', 'chrom', 'y1', 'y2', 'E']
    loop.sort_values(by=['chrom', 'x1', 'y1'])[bedpe_cols].to_csv(
        f'{output_prefix}.loop.bedpe', sep='\t', index=False, header=None)
    scloop = total_loops.loc[(total_loops['local_qval'] < fdr_thres)
                             & (total_loops['global_qval'] < fdr_thres)]
    scloop.sort_values(by=['chrom', 'x1', 'y1'])[bedpe_cols].to_csv(
        f'{output_prefix}.scloop.bedpe', sep='\t', index=False, header=None)
    bkloop = total_loops.loc[total_loops['bkfilter']]
    bkloop.sort_values(by=['chrom', 'x1', 'y1'])[bedpe_cols].to_csv(
        f'{output_prefix}.bkloop.bedpe', sep='\t', index=False, header=None)

    # find summit
    print('Finding loop summit.')
    if loop.shape[0] > 0:
        summit = pd.concat([
            find_summit(loop=sub_df,
                        res=resolution,
                        dist_thres=dist_thres // resolution)
            for chrom, sub_df in loop.groupby('chrom')
        ],
            axis=0)
        summit = summit[summit['size'] >= size_thres]
        summit.sort_values(
            by=['chrom', 'x1', 'y1'])[bedpe_cols + ['size']].to_csv(
            f'{output_prefix}.loop_summit.bedpe',
            sep='\t',
            index=False,
            header=None)
    else:
        pd.DataFrame([]).to_csv(f'{output_prefix}.loop_summit.bedpe')
    return
